"""
    A module to load configurations from environment varibles or
    YAML file(s).

"""

import os
import string
import sys
from typing import NamedTuple

import yaml

# For access to the __annotations__ attribute.
assert sys.version_info >= (3, 10), "Minimum Python 3.10 required"


class SettingsParser:
    """
    Load all settings from environment variables and/or a YAML file.

    name:
        Name of the settings. Used as the prefix for environment variables.

    yaml_file:
        YAML file to load the settings.

    first_arg:
        Treat the first argument as the settings YAML file if any.

    load_env:
        Load settings from environment variables or not.

    prefer_env:
        Override YAML vars with environment variables or vice versa.
        Default order: cmdline > YAML > Env.
        Prefer_env order: cmdline > Env. > YAML.


    Example:

    Subclass the `typing.NamedTuple`, and define a classmethod to load the
    settings.

    ```python
    class _Test(NamedTuple):
        name: str   = 'hello'

    class Settings(NamedTuple):
        YAML = None

        # Define the sections ...
        TestSettings : _Test

        @classmethod
        def load(c, yaml_file = None, first_arg = False) -> 'Settings':
            c.YAML = SettingsParser('pytest', yaml_file, first_arg=first_arg)
            return c.YAML.populate(c)

        def save(self, yaml_file = None):
            self.YAML.save(self, yaml_file=yaml_file)

    settings = Settings.load('settings.yaml')
    test = settings.TestSettings
    print(test.name)
    ```
    """

    def __init__(self, name : str, yaml_file : str = None,
            first_arg : bool = True, load_env : bool = True,
            prefer_env : bool = False):

        # Prefix of the env vars.
        # So they can be specified as `NAME_SECTION_VAR=value`
        self._name = name

        # Container for the environment variables.
        self._env  = os.environ if load_env else {}

        # Override file vars with env vars.
        self._prefer_env = prefer_env

        # Commandline args.
        self._args = self._get_cmdline_subs()
        self._pos_args = self._get_positional_args()

        # Loaded file path
        self._file = yaml_file if yaml_file else 'settings.yaml'

        # YAML file loaded variables.
        self._yamlvars : dict[str, dict[str, any]] = {}

        self.load_file(self._file, first_arg=first_arg)


    def __repr__(self) -> str:
        s = self.__class__.__name__ + ": "
        for k, v in self._yamlvars.items():
            s += f"{k} {v._asdict()}"
        return s


    def _populate_field(self, section : str, field : str, data_type, def_value):
        # Start with the default value.
        value = def_value

        # If _name is given, prefix it for the env var.
        # Format: NAME_SECTION_FIELD=value
        env_var_name = f"{section.upper()}_{field.upper()}"
        if self._name:
            env_var_name = f"{self._name.upper()}_{env_var_name}"

        # Use the Env var first, if not preferred.
        if not self._prefer_env:
            value = self._get_env(env_var_name, value)

        # Override with the YAML vars.
        value = self._get_yaml(section, field, value)

        # Override with env vars if its preferred.
        if self._prefer_env:
            value = self._get_env(env_var_name, value)

        # Check if there is any $ arg template in env vars / yaml.
        value = self._get_arg(field, value)

        # Convert to expected type.
        if value is not None:
            try:
                value = data_type(value)
            except:
                raise ValueError(f"Invalid type for {field}: {value}")
        return value


    def _populate_section(self, section_name : str, section_cls : NamedTuple):
        """ Populate settings to the current class/section. """

        assert hasattr(section_cls, '_fields'), "Section must be a NamedTuple."

        fields = {}

        # for each namedtuple fields ...
        for field in section_cls._fields:
            data_type = section_cls.__annotations__[field]

            # Default value.
            value = section_cls._field_defaults.get(field, None)

            # Populate field
            fields[field] = \
                self._populate_field(section_name, field, data_type, value)

        # Return initialized class.
        return section_cls(**fields)


    def populate(self, settings : NamedTuple):
        """ Populate settings to all sections defined by NamedTuple fields. """

        assert hasattr(settings, '_fields'), "Settings must be a NamedTuple."

        sections = {}

        # for each settings sections ...
        for section_name in settings._fields:
            section_definition = settings.__annotations__[section_name]

            # populate individual section
            value = self._populate_section(section_name, section_definition)

            # Add to class sections
            sections[section_name] = value

        # Return initialized class.
        return settings(**sections)


    def _get_positional_args(self) -> list[str]:
        positionals = []
        for i in range(1, len(sys.argv)):
            field = sys.argv[i]
            if not field.startswith("-"):
                if not sys.argv[i-1].startswith("-"):
                    positionals.append(field)
        return positionals


    def _get_cmdline_subs(self) -> dict:
        """ Create a dictionary of commandline --key value pairs.
        """
        d = {}
        i = 1
        while i < len(sys.argv):
            field = sys.argv[i]
            if field.startswith('--'):
                field = field[2:]   # remove -- prefix
                if i + 1 < len(sys.argv):
                    value = sys.argv[i+1]
                    if value.startswith('--'):
                        value = 1   # boolean flag
                else:
                    value = 1       # boolean flag
                d[field] = value
            i += 1
        return d


    def _get_arg(self, fieldname : str, value) -> str:
        """ Perform string substitution using commandline arguments.
            Substitution placeholders are prefixed with a $.
            Raises ValueError if arguments do not contain a placeholder.
        """
        if type(value) == str and '$' in value:
            tmpl = string.Template(value)
            try:
                value = tmpl.substitute(self._args)
            except KeyError as err:
                err = str(err)[1:-1] # remove quotes
                raise ValueError(
                    f"Argument required for {fieldname}='{value}', --{err} ?")
        return value


    def _get_env(self, var_name, default_value):
        return self._env.get(var_name, default_value)


    def _get_yaml_section(self, classname, default_value):
        return self._yamlvars.get(classname, default_value)


    def _get_yaml(self, classname, var_name, default_value):
        if classname in self._yamlvars:
            return self._yamlvars[classname].get(var_name, default_value)
        else:
            return default_value


    def load_file(
        self, yaml_file : str = None, first_arg : bool = False,
        update : bool = True):

        """ Load all sections and variables of a YAML file.
        Can be called multiple times to update/add over previous values.

        yaml_file:
            YAML file to load the settings.

        first_arg:
            Treat the first argument as the settings YAML file if any.

        update:
            Update the existing sections with the variables defined by the new
            file. If `False`, all existing sections will be removed.
        """

        self._file = yaml_file if yaml_file else self._file
        self._yamlvars = self._yamlvars if update else {}

        # Treat the first argument as the settings file if any.
        if first_arg and len(self._pos_args):
            self._file = self._pos_args[0]

        if os.path.isfile(self._file):
            contents = yaml.safe_load(open(self._file))
            for section, fields in contents.items():
                for fieldname, value in fields.items():
                    if section not in self._yamlvars:
                        self._yamlvars[section] = {}
                    self._yamlvars[section][fieldname] = value


    def is_loaded(self) -> bool:
        """ Returns True if at least one section from YAML file was loaded. """
        return len(self._yamlvars) > 0


    def save(self, settings : NamedTuple, yaml_file : str = None,
                keep_existing : bool = True):
        """ Save all sections of the settings to a YAML file.
            If no yaml_file is given, the initial file is used.

            keep_existing:
                Keep the already existing sections in the YAML file.
        """

        outfile = yaml_file if yaml_file else self._file

        existing = yaml.safe_load(open(outfile)) \
            if os.path.isfile(outfile) else {}

        configs = existing if keep_existing else {}

        assert type(configs) == dict, f"Existing YAML is not a dict: {outfile}"

        assert hasattr(settings, '_fields'), "Settings must be a NamedTuple."

        # for each settings sections ...
        for section_name in settings._fields:
            section = getattr(settings, section_name)

            if hasattr(section, '_asdict'):
                configs[section_name] = section._asdict()

        yaml.safe_dump(configs, open(outfile, 'w'), indent=4)
        print("Save OK:", outfile)


    def copy_sample(self, sample_file : str):
        """ Copy a sample YAML file to the current settings file. """
        import shutil
        shutil.copyfile(sample_file, self._file)
        print("Save OK:", self._file)



from pylogg.settings import NamedTuple, SettingsParser


def test_load_settings(assets, tmp_path):
    asset_file = assets / "settings.yaml"
    test_output = tmp_path / "settings.yaml"

    class Test(NamedTuple):
        row1: float = 100.0
        row2: str   = 'Package'
        row3: str   = 'Settings'
        row4: str   = 'Default'

    class Person(NamedTuple):
        name : str = 'John'
        age : int = 3

    class Settings(NamedTuple):
        YAML = None
        TestSettings : Test
        PersonSettings : Person

        @classmethod
        def load(c, yaml_file = None, first_arg = False) -> 'Settings':
            c.YAML = SettingsParser('pytest', yaml_file, first_arg=first_arg)
            return c.YAML.populate(c)

        def create(self, newfile = None):
            self.YAML.save(self, yaml_file=newfile)


    sett = Settings.load(asset_file)
    test = sett.TestSettings
    assert test.row1 == 23.6
    assert test.row2 == "Hello"
    assert test.row3 == "World"
    assert test.row4 == "Default"

    sett = sett._replace(TestSettings = sett.TestSettings._replace(row1 = 90.0))
    test = sett.TestSettings
    assert test.row1 == 90.0
    assert test.row4 == "Default"

    person = sett.PersonSettings
    assert person.name == "Mike"
    assert person.age == 29

    sett.create(test_output)
    assert 'row1: 90.0' in test_output.read_text()
    assert 'age: 29' in test_output.read_text()
    assert 'row4: Default' in test_output.read_text()


def test_yaml_write(assets, tmp_path):
    asset_file = assets / "settings.yaml"
    test_output = tmp_path / "settings.yaml"

    class Test(NamedTuple):
        row1: float = 23.6
        row2: str   = 'Hello'
        row3: str   = 'World'

    class Settings(NamedTuple):
        YAML = None
        TestSettings : Test

        @classmethod
        def load(c, yaml_file = None, first_arg = False) -> 'Settings':
            c.YAML = SettingsParser('pytest', yaml_file, first_arg=first_arg)
            return c.YAML.populate(c)

        def save(self, newfile = None):
            self.YAML.save(self, yaml_file=newfile)

    settings = Settings.load(asset_file)
    test = settings.TestSettings

    assert test.row1 == 23.6
    assert type(test.row1) == float
    assert test.row2 == 'Hello'

    settings.save(newfile=test_output)
    assert str(test.row1) in test_output.read_text()
    assert test.row2 in test_output.read_text()
    assert test.row3 in test_output.read_text()


def test_args_subs():
    import sys
    sys.argv += ['--name', 'world', '--debug', '--num', '22']

    class Test(NamedTuple):
        greeting: str   = 'Hello $name'
        number : int    = '$num'
        debug : bool    = '$debug'

    class Settings(NamedTuple):
        YAML = None
        TestSettings : Test

        @classmethod
        def load(c, yaml_file = None, first_arg = False) -> 'Settings':
            c.YAML = SettingsParser('pytest', yaml_file, first_arg=first_arg)
            return c.YAML.populate(c)

        def save(self, newfile = None):
            self.YAML.save(self, yamlfile=newfile)

    settings = Settings.load(None)
    test = settings.TestSettings

    assert test.greeting == 'Hello world'
    assert test.number == 22
    assert test.debug == True

    print(test)


def test_postitional_args():
    import sys

    yaml = SettingsParser('pytest')
    print(yaml._pos_args)

    sys.argv += ['--name', 'world', 'settings2.yaml', '--debug', '--num', '22']
    yaml = SettingsParser('pytest')
    print(yaml._pos_args)

    sys.argv += ['settings2.yaml']
    yaml = SettingsParser('pytest')
    assert yaml._pos_args == ['settings2.yaml', 'settings2.yaml']
    print(yaml._pos_args)



""" Read and write JSON and JSONL files. """

import json

import pylogg

log = pylogg.get('jsonfs')

def load_lines(jsonl_file : str) -> list[dict]:
    """ Load a JSONL file into list of dictionaries. """
    log.trace("Load JSONL:", jsonl_file)
    with open(jsonl_file) as fp:
        jsonlines = list(fp)

    return [ json.loads(json_str) for json_str in jsonlines ]


def save_lines(linelist : list[dict] , jsonl_file : str):
    """ Save a list of dictionaries as a JSONL file. """

    assert type(linelist) == list
    log.trace("Save JSONL:", jsonl_file)

    # Make sure all dict items have the same keys.
    keys = set()
    for line in linelist:
        for k in line.keys():
            keys.add(k)

    # Add the missing ones with None.
    for line in linelist:
        for k in keys:
            if k not in line:
                line[k] = None

    jsonlines = [ json.dumps(line) + "\n" for line in linelist ]
    with open(jsonl_file, "w") as fp:
        fp.writelines(jsonlines)


def save_dict(obj : dict, filename):
    """ Save a dictionary object as JSON. """
    obj = _recurse_type(obj)

    log.trace("Save JSON: ", filename)
    try:
        json.dump(obj, open(filename, "w"), indent=4)
    except Exception as err:
        print(obj)
        print(err)


def load_dict(filename) -> dict:
    """ Load a JSON file as dict. """
    log.trace("Load JSON: ", filename)
    d = json.load(open(filename))
    return d


def _recurse_type(obj):
    if type(obj) != dict:
        obj = dict(obj)

    for k, v in obj.items():
        if hasattr(v, 'item') and callable(v.item):
            # handle numpy values
            obj[k] = v.item()
        elif type(v).__name__ == 'ndarray':
            obj[k] = [i.item() for i in v]
        elif type(v) not in [dict, list, str, int, float, type(None)]:
            obj[k] = dict(v)
        elif type(v) == dict:
            obj[k] = _recurse_type(v)
    return obj



""" Raw SQL-based Postgres DB I/O utils """

import os
from datetime import datetime
from typing import NamedTuple

import pandas as pd
import psycopg
from psycopg.rows import namedtuple_row

import pylogg

log = pylogg.get('postgres')


class Tunnel(NamedTuple):
    """ Configuration for setting up a SSH tunnel. """
    host : str          = os.environ.get('TUNNEL_HOST', None)
    port : int          = int(os.environ.get('TUNNEL_PORT', 5432))
    user : str          = os.environ.get('TUNNEL_USER', '')
    pswd : str          = os.environ.get('TUNNEL_PASS', '')


class Server(NamedTuple):
    """ Configuration for connecting to a database. """
    host : str          = os.environ.get('PGDB_HOST', "localhost")
    port : int          = int(os.environ.get('PGDB_PORT', 5432))
    name : str          = os.environ.get('PGDB_NAME', "postgres")
    user : str          = os.environ.get('PGDB_USER', "admin")
    pswd : str          = os.environ.get('PGDB_PASS', "")


def ssh_tunnel(tunnel : Tunnel, ssh_port : int = 22) -> dict:
    """ Create a SSH tunnel to a remote host.
        Returns: Local tunnel endpoint dict[host, port].
    """

    from sshtunnel import SSHTunnelForwarder

    log.trace("Creating SSH tunnel to", tunnel.host)
    proxy = SSHTunnelForwarder((tunnel.host, int(ssh_port)),
            ssh_username=tunnel.user,
            ssh_password=tunnel.pswd,
            remote_bind_address=('127.0.0.1', tunnel.port),
        )
    proxy.start()
    log.done("SSH tunnel :{} ==> {}", proxy.local_bind_port, tunnel.host)
    return dict(host="127.0.0.1", port=proxy.local_bind_port)


def connect(db : Server) -> psycopg.Connection:
    """ Connect to a specific database.
        Returns: The db connection object.
    """
    log.trace("Connecting to", db.name)

    conninfo =  f'host={db.host} '\
                f'port={db.port} '\
                f'dbname={db.name} '\
                f'user={db.user} '\
                f'password={db.pswd}'

    conn = psycopg.connect(conninfo)

    log.done("Connect to database:", db.name)
    return conn


def raw_sql(conn : psycopg.Connection, psycopg_sql_string, **kwargs) -> list:
    """ Perform SQL select query using a psycopg3 style SQL statement.
        Returns the list of selected rows.
    """
    log.trace("Executing SQL:", psycopg_sql_string)

    conn.autocommit = True
    conn.row_factory = namedtuple_row
    cur : psycopg.Cursor = conn.cursor()
    rows = cur.execute(query=psycopg_sql_string, params=kwargs).fetchall()

    return rows


def query(conn : psycopg.Connection, sql, **kwargs) -> pd.DataFrame:
    """ Query the database using raw sql.
        Return a pandas dataframe containing the results.
    """
    results = raw_sql(conn, sql, **kwargs)
    if not results:
        return pd.DataFrame()
    return pd.DataFrame.from_records(results, columns=results[0]._fields)


class Table:
    def __init__(self, _tableName, **cols) -> None:
        """ Generic table definition for creating and inserting.
            Example:
                tabl = Table("test",
                    name = "varchar NOT NULL UNIQUE",
                    age  = "int4 NOT NULL"
                )
        """
        self._tblname = _tableName
        self._columns = {
            "id": f"int4 NOT NULL GENERATED BY DEFAULT AS IDENTITY(INCREMENT BY 1 MINVALUE 1 MAXVALUE 2147483647 START 1 CACHE 1 NO CYCLE)",
            "date_added": "timestamptz NOT NULL",
        }
        self._indices = {}
        self._constraints = {"id": f"{self._tblname}_pkey PRIMARY KEY"}

        for k, v in cols.items():
            self._columns[k] = v

    def __repr__(self) -> str:
        r = f"Table {self._tblname}({', '.join([k for k in self._columns.keys()])})"
        return r

    def constraint(self, col : str, value : str):
        self._constraints[col] = value

    def index(self, col : str, ops : str = None):
        if ops: assert ops in ['varchar', 'text']
        self._indices[col] = f"{ops}_pattern_ops" if ops else ""

    def _create_sql(self) -> str:
        sql = f"CREATE TABLE IF NOT EXISTS {self._tblname} (\n\t"
        rows = [f'"{k}" {v}' for k, v in self._columns.items()]
        rows += [f'CONSTRAINT {v} ("{k}")' for k, v in self._constraints.items()]
        sql += ",\n\t".join(rows) + "\n);"
        return sql

    def _index_sql(self) -> list[str]:
        return [
            f'CREATE INDEX IF NOT EXISTS ix_{self._tblname}_{k} ON {self._tblname} USING btree ("{k}" {v});'
            for k, v in self._indices.items()
        ]

    def create_all(self, conn : psycopg.Connection, drop_existing = False):
        """ Create the table and all the defined indices.
        """
        log.trace("Creating table", self._tblname)

        try:
            if drop_existing:
                log.warn("Dropping existing table:", self._tblname)

                for k in self._indices.keys():
                    sql = f"DROP INDEX IF EXISTS ix_{self._tblname}_{k};"
                    conn.execute(sql)

                sql = f"DROP TABLE IF EXISTS {self._tblname};"
                conn.execute(sql)

            for sql in [self._create_sql()] + self._index_sql():
                log.trace(sql)
                conn.execute(sql)

            conn.commit()
            log.done("Table create:", self._tblname)

        except:
            conn.rollback()
            log.error("Failed to create table:", self._tblname)
            raise

    def insert_row(self, conn : psycopg.Connection, **kwargs):
        """ Insert a single table row and commit immediately.
            Returns the inserted Row with the id field. Example:
            tabl.insert_row(name = "John", age = 31)
        """
        log.trace("Insert row into", self._tblname)

        if "date_added" not in kwargs:
            kwargs["date_added"] = datetime.now()

        columns = [k for k in kwargs.keys()]
        for col in columns:
            assert col in self._columns, f"Column '{col}' not in {self}"

        column_string = ", ".join(columns)
        values_string = ", ".join([f"%({k})s" for k in columns])

        sql = f"""
        INSERT INTO {self._tblname} ({column_string})
        VALUES ({values_string}) RETURNING id;
        """
        row = raw_sql(conn, sql, **kwargs)

        if row:
            return row[0]
        return None

    def copy(self, conn : psycopg.Connection, rows : list[dict]):
        """ Insert a batch of rows and commit at the end. Example:
            tabl.copy([
                {'name': "Mike", 'age': 23},
                {'name': "Kate", 'age': 41},
                {'name': "Terry", 'age': 9},
            ])
        """

        t1 = log.info("Copying {} rows into {}", len(rows), self._tblname)

        columns = [k for k in rows[0].keys()]
        if "date_added" not in columns:
            columns.append('date_added')

        for col in columns:
            assert col in self._columns, f"Column '{col}' not in {self}"

        column_string = ", ".join(columns)
        sql = f"COPY {self._tblname} ({column_string}) FROM STDIN"

        cur : psycopg.Cursor = conn.cursor()
        try:
            with cur.copy(sql) as copy:
                for row in rows:
                    if 'date_added' not in row:
                        row['date_added'] = datetime.now()
                    copy.write_row([row[k] for k in columns])

            conn.commit()
            t1.done("Copy {} rows to {}", len(rows), self._tblname)
        except:
            conn.rollback()
            log.error("Failed to copy rows to", self._tblname)
            raise


if __name__ == "__main__":
    server = Server()

    if os.environ.get("TUNNEL_HOST", None):
        tunnel = Tunnel()
        server = server._replace(**ssh_tunnel(tunnel))

    tabl = Table("test",
        name   = "varchar NOT NULL UNIQUE",
        age    = "int4 NOT NULL"
    )

    tabl.index('age')
    tabl.index('name', 'varchar')

    db = connect(server)
    tabl.create_all(db, drop_existing=True)

    print(tabl.insert_row(db, name = "John", age = 31))
    print(tabl.insert_row(db, name = "Doe", age = 32))

    tabl.copy(db, [
        {'name': "Mike", 'age': 23},
        {'name': "Kate", 'age': 41},
        {'name': "Terry", 'age': 9},
    ])

    for row in raw_sql(db, "SELECT * FROM test;"):
        print(row)

    print("Done")
